# UV_for_Bldg_03                                                     By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
# 最初に、OSCより取得したforestデータ(Forest)とXPの地形メッシュ(MESH)を準備します。
# Forstに間の点を追加し、指定されたズレをランダムに不可します。祖て多店軍を地形メッシュに
# Shrinkwrapし、原点を地形表面に移動しsます。
# 詳細は XP11用シーナリー作成メモ(拡張偏)03版以降 の　項を参照。
#
import bpy
import math
import copy
import bmesh
import mathutils
import datetime
import numpy as np
import random
from mathutils import Matrix,Vector
#
####### 三角形の内外判定 ############
def is_in_triangle(a, b, c, p):
  v_a = b - a
  v_b = c - a
  v_p = p - a
  A = np.concatenate([v_a, v_b], axis=1)
  w = np.linalg.inv(A) * v_p
  if np.all(w>=0)and np.all(w<=1) and w.sum() <= 1.0:
    return True
  else:
    return False
#
####### 地形メッシュ(MESH)の原点(0,0)での高さ terrz 算出
def elevation_at_center(mesh_name, xpt, ypt):
  p = np.matrix([[xpt],[ypt]])                  # 高さを求める位置Pを(0,0)とする
  aa = bpy.data.objects[mesh_name]              # "MESH"オブジェクトをaaと呼ぶ
  bpy.context.view_layer.objects.active = aa    # そのオブジェクトをActiveに
  bpy.ops.object.mode_set(mode='EDIT')          # 編集モ－ドにする
  bm = bmesh.from_edit_mesh(aa.data)
  bm.faces.ensure_lookup_table()
  #
  for f in range(len(bm.faces)):
    ax = bm.faces[f].verts[0].co.x
    ay = bm.faces[f].verts[0].co.y
    az = bm.faces[f].verts[0].co.z
    bx = bm.faces[f].verts[1].co.x
    by = bm.faces[f].verts[1].co.y
    bz = bm.faces[f].verts[1].co.z
    cx = bm.faces[f].verts[2].co.x
    cy = bm.faces[f].verts[2].co.y
    cz = bm.faces[f].verts[2].co.z
    a = np.matrix([[ax],[ay]])
    b = np.matrix([[bx],[by]])
    c = np.matrix([[cx],[cy]])
    spotted = is_in_triangle(a, b, c, p)
    if spotted:
      k1 = ((by-ay)*(cz-az)-(cy-ay)*(bz-az))*(0-ax)
      k2 = ((bz-az)*(cx-ax)-(cz-az)*(bx-ax))*(0-ay)
      k3 = (cx-ax)*(by-ay)-(bx-ax)*(cy-ay)
      terrz = (k1 + k2) / k3 + az          # (0,0)原点での地表のZ座標を求める
      
      break
  return terrz
#
#######　"MESH"の(0,0)の高さを 0 とする　###################################
obj = bpy.data.objects["MESH"]            # Tree_posをobj1と呼ぶ
obj.select_set(True)
bpy.ops.object.mode_set(mode='EDIT')      # 編集モ－ドにする
meshdata = bmesh.from_edit_mesh(obj.data)# メッシュデータ作成
terrz = elevation_at_center("MESH", 0, 0) # (0,0)での地表の高さをterrzとする
for v in meshdata.verts:
  v.co.z -= terrz                         # 建物各頂点を下にterrz移動
bpy.ops.object.mode_set(mode='OBJECT')    # オブジェクトモードにする
#
now = datetime.datetime.now()
print("Elevation of MESH at(0,0) =", terrz)
print("Elevatioin corrected.", now )