# Sorting_by_height                                              By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
#
# OSMより抽出された建物objで、垂直線の高さが6.2m未満のものを Bldg-S1 または Bldg-S2、
# 6.2m以上33m未満のものを Bldg-M1 または Bldg-M2、33m以上のものを Bldg-L のグループ
# に分けます。もし Bldg-S 又は Bldg-M の建物数が約1万戸より多ければ Bldg-M1 や Bldg-M2の様に末尾に
# 数字を付けて更に分けます。詳細は XP用OSMシーナリー作成　02版以降参照。
#
import bpy
import bmesh
#
#########  メインプログラム 
name_ob = bpy.context.object.name              # 対象obj名をname_obに保存
bpy.context.object.name = "Bldg-S"             # Activeなobj名を"Bldg-S"とします
a= bpy.context.object                          # そのobjを"a"として扱います
bpy.ops.object.mode_set(mode='EDIT')           # 編集モ－ドにする
bpy.ops.mesh.select_all(action='DESELECT')     # メッシュの点を全て非選択に
bm = bmesh.from_edit_mesh(a.data)
for e in bm.edges:
    x0 = e.verts[0].co.x
    y0 = e.verts[0].co.y
    x1 = e.verts[1].co.x
    y1 = e.verts[1].co.y
    if x0 == x1 and y0 == y1:                  # 線が垂直の場合で
        if e.calc_length() > 6.2:              # 線の高さ6.2m以上の場合
            e.select_set(True)                 # その線を選択する
bpy.ops.mesh.select_linked(delimit=set())      # その線を含む建物メッシュを選択する
bpy.ops.mesh.separate(type='SELECTED')         # それら建物を元のobjと別のobjにする
bpy.ops.object.mode_set(mode='OBJECT')         # オブジェクトモードにする
for obj in bpy.context.selected_objects:
    if obj.name != "Bldg-S":                   # obj名が"Bldg_S"でなければ
        obj.name="Bldg-M"                      # そのobj名を"Bldg-M"とする
        obj.select_set(True)                   # そのobjを選択し
        bpy.context.view_layer.objects.active = obj    # そのobjをActiveにする
#
b= bpy.context.object                          # そのobjを以下 b として扱う
bpy.ops.object.mode_set(mode='EDIT')           # 編集モ－ドにする
bbm = bmesh.from_edit_mesh(b.data)             # そのobjのメッシュをbbmとします
sel_count=0
for e in bbm.edges:
    x0 = e.verts[0].co.x
    y0 = e.verts[0].co.y
    x1 = e.verts[1].co.x
    y1 = e.verts[1].co.y
    if x0 == x1 and y0 == y1:                  # bbmの線が垂直の場合で
        if e.calc_length() > 33:               # 線の高さ33m以上の場合
            e.select_set(True)                 # その線を選択する
            sel_count += 1
if sel_count != 0:
    bpy.ops.mesh.select_linked(delimit=set())      # その線を含む建物全てを選択する
    bpy.ops.mesh.separate(type='SELECTED')         # それら建物を元のobjと別のobjにする
    bpy.ops.object.mode_set(mode='OBJECT')         # オブジェクトモ－ドにする
    for obj in bpy.context.selected_objects:
        if obj.name != "Bldg-S" and obj.name != "Bldg-M":
            obj.name="Bldg-H"                      # そのobj名を"Bldg-H"とする
            obj.select_set(True)                   # そのobjを選択し
#bpy.ops.object.mode_set(mode='OBJECT')         # オブジェクトモ－ドに戻す
#
#########  必要ならBldg-M, Bldg-S を更に分割
for ob in bpy.context.scene.objects:
    ob.select_set(False)                       #全てのOBJを非選択にする
check_obj = bpy.data.objects['Bldg-M']
check_obj.select_set(True)                     #そのOBJを選択にする
bpy.context.view_layer.objects.active = check_obj
me = check_obj.data
if len(me.polygons) > 100000:
    bpy.context.object.name = "Bldg-M1"        #元のオブジェクト名に"1"を追加する
    bpy.ops.object.mode_set(mode='EDIT')       #編集モ－ドにする
    meshdata = bmesh.from_edit_mesh(me)
    meshdata.select_mode = {'VERT'}            #選択モードを頂点選択モードにする
    bpy.ops.mesh.select_random(percent=4.5, seed=4)  #点の4.5%をランダムに選択する
    bpy.ops.mesh.select_linked(delimit=set())  #その点に接続する部分を選択する
    bpy.ops.mesh.separate(type='SELECTED')     #選択された部分を別オブジェクトにする
    bpy.ops.object.mode_set(mode='OBJECT')     #オブジェクトモードにする
    bpy.data.objects["Bldg-M1.001"].name = "Bldg-M2"
for ob in bpy.context.scene.objects:
    ob.select_set(False)                       #全てのOBJを非選択にする
check2_obj = bpy.data.objects['Bldg-S']
check2_obj.select_set(True)                     #そのOBJを選択にする
bpy.context.view_layer.objects.active = check2_obj
me2 = check2_obj.data
if len(me2.polygons) > 100000:
    bpy.data.objects["Bldg-S"].name = "Bldg-S1"
    bpy.ops.object.mode_set(mode='EDIT')       #編集モ－ドにする
    meshdata2 = bmesh.from_edit_mesh(me2)
    meshdata2.select_mode = {'VERT'}            #選択モードを頂点選択モードにする
    bpy.ops.mesh.select_random(percent=4.5, seed=4)  #点の4.5%をランダムに選択する
    bpy.ops.mesh.select_linked(delimit=set())  #その点に接続する部分を選択する
    bpy.ops.mesh.separate(type='SELECTED')     #選択された部分を別オブジェクトにする
    bpy.ops.object.mode_set(mode='OBJECT')     #オブジェクトモードにする
    bpy.data.objects["Bldg-S1.001"].name = "Bldg-S2"
print(name_ob, "sorting finished.")