# Tree_plant_01                                                     By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
# 最初に、(1)B-leaf_pos, C-leaf_posに樹木を置く位置に点を配置します。
#　　　　　　(2)樹木を配置する空オブジェクト"Leaves_A","Leaves_B","Leaves_C"を作成します。
# 実行すると、樹木の位置に広葉樹、針葉樹がランダムに選択されて配置されます。
# 詳細は XP11用シーナリー作成メモ(拡張偏)03版以降 の5項を参照。
#
import bpy
import math
import copy
import bmesh
import mathutils
import datetime
import numpy as np
import random
from mathutils import Matrix,Vector
#
######## 編集モードの原点移動サブプログラム
def move_obj_center(x,y,z):
  obj = bpy.context.object
  new_origin = Vector((x,y,z))                  # 新しい原点座標(グローバル)
  world_loc = new_origin
  local_loc = obj.matrix_world.inverted() @ world_loc    #ーカル座標に変換
  mat = Matrix.Translation(-local_loc)          # 位置だけのマトリックスワールド
  if obj.data.is_editmode:                      # オブジェクトが編集モードの場合
    bm = bmesh.from_edit_mesh(obj.data)
    bm.transform(mat)                           # メッシュを移動
    bmesh.update_edit_mesh(obj.data, False, False)
  else:                                         # オブジェクトモードの場合
    obj.data.transform(mat)                     # メッシュを移動
  obj.data.update()
  obj.matrix_world.translation = world_loc      # オブジェクト自体を移動
  return
#
####### "Tree_pos"に樹木各部を配置し、LOD設定毎にobj作成　#########  
def Tree_plant(target_name,part_name,xp,yp,zp,ang):
  obj1 = bpy.data.objects[ part_name ]          #追加元をobj1と呼ぶ
  obj1.select_set(True)                         #obj1を選択
  bpy.context.view_layer.objects.active = obj1  # Activeにする
  bpy.ops.object.duplicate()                    #obj1をコピーする
  obj1.select_set(False)                        #追加元の選択は解除する
  bpy.ops.transform.rotate(value=ang, orient_axis='Z')#obj1を回転する
  bpy.ops.transform.translate(value=(xp,yp,zp)) #obj1を移動する
  move_obj_center(0,0,0)                        #obj1の原点を(0,0,0)にする
  obj2 = bpy.data.objects[ target_name ]        #移動先をobj2とする
  obj2.select_set(True)                         #広葉樹のそのタイプを選択
  bpy.context.view_layer.objects.active = obj2  # Activeにする
  bpy.ops.object.join()                         #移動先に結合する
  obj2.select_set(False)                        #移動先の選択を解除する
  return obj2
#
######　 パラメータ設定　#############################################
Btype = ["B-leaf1","B-leaf2","B-leaf3"]  # 広葉樹(Brosd-leaf)の種類指定
Bweight = [0.3, 0.4, 0.3]                # 広葉樹の頻度指定
Ctype = ["C-leaf1","C-leaf2"]            # 針葉樹(Conifer)の種類指定
Cweight = [0.5, 0.5]                     # 針葉樹の頻度指定
Angshift = [-0.628, -0.314, 0, 0.314, 0.628] #回転ズレ
Angweight = [0.2, 0.2, 0.2, 0.2, 0.2]  #回転ズレ頻度
#
######　メインプログラム　#############################################Tree
now = datetime.datetime.now()
print()
print("Tree planting starts at", now)      # 開始時刻を印刷
for obj in bpy.context.selected_objects:
  obj.select_set(False)                    # 一旦オブジェクトを全て非選択にする。
#
######　広葉樹の樹木を作成　#######
ob = bpy.data.objects.get("Broad-leaf_pos")
if ob != None:
  obj1 = bpy.data.objects["Broad-leaf_pos"]      # Tree_posをobj1と呼ぶ
  obj1.select_set(True)
  bpy.ops.object.mode_set(mode='EDIT')       # 編集モ－ドにする
  mesh1 = bmesh.from_edit_mesh(obj1.data)    # メッシュデータ作成
  Xp = [v.co.x for v in mesh1.verts]
  Yp = [v.co.y for v in mesh1.verts]
  Zp = [v.co.z for v in mesh1.verts]
  bpy.ops.object.editmode_toggle()           # オブジェクトモ－ドにする
  obj1.select_set(False)
  print("Total Broad-leaf trees =", len(Xp))
  for nn in range(0,len(Xp)):
    print("\r", nn, end="")
    Bsel = random.choices(Btype, weights = Bweight, k=1) #広葉樹の選択
    Ang = random.choices(Angshift, weights = Angweight, k=1)
    Tree_plant("Leaves_A",Bsel[0]+"A",Xp[nn],Yp[nn],Zp[nn],Ang[0]) 
    Tree_plant("Leaves_B",Bsel[0]+"A",Xp[nn],Yp[nn],Zp[nn],Ang[0]+1.571) 
    Tree_plant("Leaves_C",Bsel[0]+"B",Xp[nn],Yp[nn],Zp[nn],Ang[0]) 
#
######　針葉樹の樹木を作成　#######
ob = bpy.data.objects.get("Conifer_pos")
if ob != None:
  obj1 = bpy.data.objects["Conifer_pos"]      # Tree_posをobj1と呼ぶ
  obj1.select_set(True)
  bpy.ops.object.mode_set(mode='EDIT')       # 編集モ－ドにする
  mesh2 = bmesh.from_edit_mesh(obj1.data)    # メッシュデータ作成
  Xp = [v.co.x for v in mesh2.verts]
  Yp = [v.co.y for v in mesh2.verts]
  Zp = [v.co.z for v in mesh2.verts]
  bpy.ops.object.editmode_toggle()           # オブジェクトモ－ドにする
  obj1.select_set(False)
  print("Total Conifer trees =", len(Xp))
  for nn in range(0,len(Xp)):
    print("\r", nn, end="")
    Csel = random.choices(Ctype, weights = Cweight, k=1) #広葉樹の選択
    Ang = random.choices(Angshift, weights = Angweight, k=1)
    Tree_plant("Leaves_A",Csel[0]+"A",Xp[nn],Yp[nn],Zp[nn],Ang[0]) 
    Tree_plant("Leaves_B",Csel[0]+"A",Xp[nn],Yp[nn],Zp[nn],Ang[0]+1.571) 
    Tree_plant("Leaves_C",Csel[0]+"B",Xp[nn],Yp[nn],Zp[nn],Ang[0]) 
#
now = datetime.datetime.now()
print()
print("All Tree placed at", now )