# Tree_pos_01                                                     By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
# 最初に、XPの地形メッシュ(MESH)を準備します。地形メッシュ上にOSCよりforestデータ(Forest)を作成します。
# 本スクリプトにより10m格子の点を抽出し、とその中間の位置に点を追加します。さらに各点をランダムにずらし自然に見える様に、
# 樹木配置点を作成しますします。
# 詳細は XP11用OSMシーナリー作成02版以降 の5項を参照。
#
import bpy
import math
import copy
import bmesh
import mathutils
import datetime
import numpy as np
import random
from mathutils import Matrix,Vector
#
######　 パラメータ設定　#############################################
pos_weight = [0.05, 0.12, 0.27, 0.12, 0.27, 0.12, 0.05]   # 広葉樹の頻度指定
dist_posX = [-1.5, -1, -0.5, 0, 0.5, 1, 1.5]
dist_posY = [-1.5, -1, -0.5, 0, 0.5, 1, 1.5]
#
######　メインプログラム　#############################################
now = datetime.datetime.now()
print()
print("Tree setting starts at", now)              # 開始時刻を印刷
#
selectob = bpy.data.objects.get("Forest")
if selectob == None:
  print("Forest object missing.")                 #Forestが無い事を通知
#
obj1 = bpy.data.objects["Forest"]
obj1.select_set(True)
bpy.context.view_layer.objects.active = obj1      # Activeにする
bpy.ops.object.mode_set(mode='EDIT')              # 編集モ－ドにする
meshdata = bmesh.from_edit_mesh(obj1.data)        # メッシュデータ作成
#
Verts = []
for v in meshdata.verts:
  if v.is_boundary == False:
    PosX = random.choices(dist_posX, weights = pos_weight, k=1)
    PosY = random.choices(dist_posY, weights = pos_weight, k=1)
    Verts.append( [ v.co.x+PosX[0], v.co.y+PosY[0], 0 ] )
#
for f in meshdata.faces:
  if f.calc_area() > 80:
    f.select_set(True)
meshdata.select_mode = {'VERT'}                   #点選択モードにする
bpy.ops.mesh.select_all(action='INVERT')
bpy.ops.mesh.delete(type='VERT')
meshdata.select_mode = {'FACE'}                   #面選択モードにする
bpy.ops.mesh.select_all(action='SELECT')
#
for f in meshdata.faces:
  if f.calc_area() > 80:
    Xmean = 0
    Ymean = 0
    n = 0
    for v in f.verts:
      Xmean += v.co.x # 建物各頂点のx座標リスト
      Ymean += v.co.y # 建物各頂点のy座標リスト
      n += 1
    Xmean = Xmean / n
    Ymean = Ymean / n
    PosX = random.choices(dist_posX, weights = pos_weight, k=1)
    PosY = random.choices(dist_posY, weights = pos_weight, k=1)
    Verts.append( [ Xmean+PosX[0], Ymean+PosY[0], 0 ] )
print("Total number of trees = ", len(Verts))  #木の総数表示
#
treemesh = bpy.data.meshes.new("treemesh")   #treemeshというデータを作る
treemesh.from_pydata(Verts,[],[])            #そのデータは点群(Verts)のみとする
treemesh.update()
obj2 = bpy.data.objects.new("Broad-leaf_pos", treemesh)#Treeと云うobjを作る
bpy.context.collection.objects.link(obj2)    #そのobjをシーンに加える
obj2.select_set(True)                        #Treeを選択する
bpy.context.view_layer.objects.active = obj2
obj1.select_set(False)                       #Forestを非選択にする
#
now = datetime.datetime.now()
print("All processing done at", now)