# UV_for_Bldg_02                                                     By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
#
# 選択中のオブジェクト（Build-H, Build-M,Build-S のみ） について建物の屋根を修正し、
# 規定のルールでテクスチャを貼ります。最後に建物のWorld原点を地表に上下移動します。
# 詳細は XP11用シーナリー作成メモ(拡張偏)03版以降 の6項を参照。
#
import bpy
import math
import copy
import bmesh
import mathutils
import datetime
import numpy as np
import random
from mathutils import Matrix,Vector
#
######## 編集モードの原点移動サブプログラム
def move_obj_center(x,y,z):
  obj = bpy.context.object
  new_origin = Vector((x,y,z))                  # 新しい原点座標(グローバル)
  world_loc = new_origin
  local_loc = obj.matrix_world.inverted() @ world_loc    #ーカル座標に変換
  mat = Matrix.Translation(-local_loc)          # 位置だけのマトリックスワールド
  if obj.data.is_editmode:                      # オブジェクトが編集モードの場合
    bm = bmesh.from_edit_mesh(obj.data)
    bm.transform(mat)                           # メッシュを移動
    bmesh.update_edit_mesh(obj.data, False, False)
  else:                                         # オブジェクトモードの場合
    obj.data.transform(mat)                     # メッシュを移動
  obj.data.update()
  obj.matrix_world.translation = world_loc      # オブジェクト自体を移動
  return
#
####### 三角形の内外判定 ############
def is_in_triangle(a, b, c, p):
  v_a = b - a
  v_b = c - a
  v_p = p - a
  A = np.concatenate([v_a, v_b], axis=1)
  w = np.linalg.inv(A) * v_p
  if np.all(w>=0)and np.all(w<=1) and w.sum() <= 1.0:
    return True
  else:
    return False
#
#########################################################################
name_obj = bpy.context.object.name
name_obj_main = name_obj[0:6]        # 先頭6文字をname_obj_main と呼ぶ
if name_obj_main == "Bldg-H":        # Bldg-H のテクスチャ画像情報
  UV_NAME ="UVMap"                   # UVを書き込むUVマップ名称、ディフォルトは "UVMap"
  TEX_SIZE_H = 8192                  # 画像の水平サイズ(dot)
  TEX_SIZE_V = 512                   # 画像の垂直サイズ(dot)
  #
  ######### 屋根テクスチャパラメーター
  #                 [0]     [1]    [2]    [3]   [4]   [5]    [6]
  #                H解像    V解像  　H原点   V原点　 左端　  右端    重み
  #               (dot/m) (dot/m) (dot)  (dot) (dot) (dot)
  RP = np.array([   10,     10,     0,     0,    0,  1030,   1.0])#パターン0
  #
  ######### 壁テクスチャパラメーター
  #                [0]     [1]   [2]   [3]   [4]   [5]   [6]   [7]   [8]
  #               H解像　  V解像   H原点  V原点 H周期  V周期  左余白 上余白  重み
  #              (dot/m)(dot/m) (dot) (dot) (dot) (dot) (dot) (dot)
  WP0 = np.array([ 8.2 ,  8.2 , 1030,   0,   34,   32,    0,    8,  0.15])#0
  WP1 = np.array([ 8.0 ,  8.0 , 2050,   0,   31,   32,    0,   15,  0.15])#1
  WP2 = np.array([ 8.0 ,  8.0 , 3086,   0,   45,   32,   12,   10,  0.15])#2
  WP3 = np.array([ 8.0 ,  8.0 , 4095,   0,   45,   32,    0,   13,  0.15])#3
  WP4 = np.array([ 8.0 ,  8.0 , 5119,   0,   60,   32,    4,    7,  0.15])#4
  WP5 = np.array([ 8.0 ,  8.0 , 6144,   0,   53,   32,   11,   20,  0.1 ])#5
  WP6 = np.array([ 8.75,  8.75, 7168,   0,   64,   35,    0,    0,  0.15])#6
  WP = np.vstack([WP0, WP1, WP2, WP3, WP4, WP5, WP6])
  Wnum= 7
##########################################################################
elif name_obj_main == "Bldg-M":       # Bldg-M のテクスチャ画像情報
  UV_NAME ="UVMap"                    # UVを書き込むUVマップ名称、ディフォルトは "UVMap"
  TEX_SIZE_H = 512                    # 画像の水平サイズ(dot)
  TEX_SIZE_V = 8192                   # 画像の垂直サイズ(dot)
  #
  ######### 屋根テクスチャパラメーター
  #                [0]     [1]    [2]   [3]   [4]   [5]
  #                H解像   V解像   H原点  V原点  上限V 下限V
  #              (dot/m) (dot/m) (dot) (dot) (dot) (dot)
  RP = np.array([   10,     10,    0,  7676, 8192, 7160 ])
  #
  ######### 壁テクスチャパラメーター
  #              [0]     [1]  [2]  [3]    [4]   [5]   [6]  [7]  [8]
  #              H解像  V解像  H原点 V原点  H周期  V周期   左余 上余  重み
  #              dot/m  dot/m dot  dot    dot   dit   dot  dot
  WP0 =np.array([  9.7 , 9.7 ,  0, 6819, 34.1,   34,   17,   8,  0.1  ]) #0
  WP1 =np.array([ 16.28,16.28,  0, 6394,   32,   57,   13,  15,  0.1  ]) #1
  WP2 =np.array([ 12.28,12.28,  0, 5921,   64,   43,    6,  26,  0.05 ]) #2
  WP3 =np.array([ 15.7 ,15.7 ,  0, 5316,   64,   55,    2,   3,  0.1  ]) #3
  WP4 =np.array([     0,    0,  0, 4773,    0,    0,    0,   0,  0.0  ]) #4
  WP5 =np.array([ 15.4 ,15.4 ,  0, 4161,   64,   54,   32,  11,  0.05 ]) #5
  WP6 =np.array([ 13.4 ,13.4 ,  0, 3646,   64,   47,   18,  14,  0.1  ]) #6
  WP7 =np.array([ 14.0 ,14.0 ,  0, 3087,   64,   49,   40,  18,  0.1  ]) #7
  WP8 =np.array([ 26.85,26.85,  0, 2367,   64,   94,    9,  27,  0.1  ]) #8
  WP9 =np.array([  6.2 , 6.2 ,  0, 1982,   64,   25,    1,   0,  0.1  ]) #9
  WP10=np.array([ 20.5 ,20.5 ,  0, 1185,  128,   72,    6,   6,  0.1  ]) #10
  WP11=np.array([ 16.28,16.28,  0,  555,  128,   57,    6,  15,  0.05 ]) #11
  WP12=np.array([ 14.28,14.28,  0,    0,   64,   50,   10,  17,  0.05 ]) #12
  WP =np.vstack([WP0,WP1,WP2,WP3,WP4,WP5,WP6,WP7,WP8,WP9,WP10,WP11,WP12])
  Wnum= 13                                                        # 壁パターン数
###########################################################################  
elif name_obj_main == "Bldg-S":       # Bldg-S のテクスチャ画像情報
  UV_NAME ="UVMap"                    # UVを書き込むUVマップ名称、ディフォルトは "UVMap"
  TEX_SIZE_H =512                     # 屋根画像の水平サイズ(dot)
  TEX_SIZE_V =8192                   # 屋根画像の垂直サイズ(dot)
  #
  ######### 屋根テクスチャパラメーター
  #                 [0]   [1]   [2]   [3]   [4]   [5]   [6]
  #                H解像 V解像 H原点 V原点 上限V 下限V  重み
  #               dot/m  dot/m  dot   dot   dot    dot
  RP0 = np.array([  50,    50,   0,  8186,  8186, 7796, 0.1 ]) #0
  RP1 = np.array([  50,    50,   0,  7786,  7786, 7396, 0.1 ]) #1
  RP2 = np.array([  50,    50,   0,  7387,  7387, 6997, 0.15]) #2
  RP3 = np.array([  50,    50,   0,  6987,  6987, 6597, 0.15]) #3
  RP4 = np.array([  50,    50,   0,  6592,  6592, 6202, 0.2 ]) #4
  RP5 = np.array([  50,    50,   0,  6192,  6192, 5802, 0.3 ]) #5
  RP6 = np.array([  50,    50,   0,  5792,  4091, 4096, 0.0 ]) #6
  RP7 = np.array([  50,    50,   0,  5192,  5192, 4802, 0.0 ]) #7
  RP = np.vstack([RP0, RP1, RP2, RP3, RP4, RP5, RP6, RP7 ])
  Rnum= 8                                                  # 屋根パターン数
  #
  ######### 壁テクスチャパラメーター
  #                [0]    [1]  [2]   [3]  [4]    [5]  [6]  [7]  [8]
  #               H解像  V解像  H原点  V原点 H周期  V周期 左余  上余  重み
  #              dot/m  dot/m  dot   dot  dot    dot  dot  dot
  WP0 = np.array([71.1, 35.3,   0,  4752, 128,   106,  6,   0,  0.1 ]) #0
  WP1 = np.array([71.1, 35.0,   0,  4492, 128,   105,  9,   0,  0.1 ]) #1
  WP2 = np.array([71.1, 38.6,   0,  4232, 128,   116,  6,   0,  0.1 ]) #2
  WP3 = np.array([40.6, 32.0,   0,  3952, 256,   113,  8,  48,  0.1 ]) #3
  WP4 = np.array([28.4, 36.6,   0,  3692, 128 ,  107,  5,  15,  0.1 ]) #4
  WP5 = np.array([71.1, 34.7,   0,  3432, 128,   104,  6,   0,  0.1 ]) #5
  WP6 = np.array([71.1, 35.0,   0,  3152, 128,   105, 10,   0,  0.1 ]) #6
  WP7 = np.array([40.6, 44.3,   0,  2862, 256,   133, 11,   0,  0.1 ]) #7
  WP8 = np.array([40.6, 44.3,   0,  2548, 256,   133, 17,   0,  0.1 ]) #8
  WP9 = np.array([40.6, 44.3,   0,  2241, 256,   133, 17,   0,  0.1 ]) #9
  WP = np.vstack([WP0, WP1, WP2, WP3, WP4, WP5, WP6, WP7, WP8, WP9])
  Wnum= 10
#
now = datetime.datetime.now()
print()
print(name_obj, "separation starts at", now)   # 開始時刻を印刷
bpy.ops.object.mode_set(mode='EDIT')           # 編集モ－ドにする
bpy.ops.mesh.separate(type='LOOSE')            # 離れた部分を別objにする
bpy.ops.object.mode_set(mode='OBJECT')         # オブジェクトモードにする
now = datetime.datetime.now()
print("UV setting starts at", now)             # 開始時刻を印刷
selected_name_list = [obj.name for obj in bpy.context.selected_objects]
print("Total number of objects:", len(selected_name_list))   # 対象Obj数印刷
for obj in bpy.context.selected_objects:
  obj.select_set(False)                        # 一旦オブジェクトを全て非選択にする。
#
####### 以下、個々の建物について形状を認識しUV設定を行う  ############################
nn = 1
for obj_name in selected_name_list:
  obj = bpy.data.objects[obj_name]             # 対象建物をobjと呼ぶ
  obj.select_set(True)
  bpy.context.view_layer.objects.active = obj  # その建物をActiveにする
  if obj.type != 'MESH':
    print("WARNING: Obj is not MESH.")
    break                                      # その建物がメッシュでない場合は処理しない
  #
  ####### 建物の高さ等基本形状取得  ################################
  Zlist = []
  bpy.ops.object.mode_set(mode='EDIT')         # 編集モ－ドにする
  meshdata = bmesh.from_edit_mesh(obj.data)    # メッシュデータ作成
  for v in meshdata.verts: Zlist.append(v.co.z)# 建物各頂点のZ座標リスト
  Ztop = max(Zlist)                            # 棟(むね)のZ座標(Ztop)算出
  Zbase = min(Zlist)                           # 基盤面のZ座標(Zbase)算出
  Htop = Ztop - Zbase                          # 屋根高さ(Htop)算出
  Zeave = []                                   # 軒高のリスト
  Zeave = [Zlist[n] for n in range(len(Zlist)) if (Zlist[n]<Ztop) and (Zlist[n]>Zbase)]
  #
  ##########  各面のMaterialで、壁は0、屋根は1 に統一する  ############
  for f in meshdata.faces:
    ncase = 0
    for v in f.verts:
      if v.co.ｚ == Zbase:
        f.material_index = 0                   # 壁の判定
        ncase = 1
        break
    if ncase == 0: f.material_index = 1        # 屋根の判定  
  #
  if Zeave == []:        # 切妻屋根になっていない建物の場合、東南角を原点として向きを算出する
    Wlong = 0
    for e in meshdata.edges:
      if e.verts[0].co.z == Zbase and e.verts[1].co.z == Zbase:
        elength = e.calc_length()
        if elength > Wlong:
          Wlong = elength              # 壁の最大幅(建物の長さ)
          X0 = e. verts[0].co.x 
          Y0 = e. verts[0].co.y
          X1 = e. verts[1].co.x
          Y1 = e. verts[1].co.y
    if X0 > X1: X0,Y0,X1,Y1 = X1,Y1,X0,Y0
    if X0 == X1:
      k = 999
      Zang = 1.57                      # 建物の角度(東を0)
    else:
      k = (Y1-Y0)/(X1-X0)              # 建物の向き(東を0)
      Zang = math.atan(k)              # 建物の角度(東を0)
  #
  else:              #  切妻屋根になっている建物で、棟の西南端を原点として建物の向きを求める
    Zroof = min(Zeave)
    for e in meshdata.edges:
      if e.verts[0].co.z == Ztop and e.verts[1].co.z == Ztop:
        X0 = e. verts[0].co.x 
        Y0 = e. verts[0].co.y
        X1 = e. verts[1].co.x
        Y1 = e. verts[1].co.y
        Wlong = e.calc_length()        # 建物の棟方向の長さを求める
        break
    if X0 > X1: X0,Y0,X1,Y1 = X1,Y1,X0,Y0
    Xctr = ( X0 + X1 ) / 2             # 棟中央のX座標
    Yctr = ( Y0 + Y1 ) / 2             # 棟中央のY座標
    if X0 == X1:
      k = 999
      Zang = 1.57                      # 建物の角度(東を0)
    else:
      k = (Y1-Y0)/(X1-X0)              # 建物の向き(東を0)
      Zang = math.atan(k)              # 建物の角度(東を0)
    #
    ########################################################################
    if name_obj_main == "Bldg-S":         # Bldg-S の時、屋根を4寸勾配にしてひさしを出す
      Wshort = 0
      for e in meshdata.edges:
        if e.verts[0].co.z == Zbase and e.verts[1].co.z == Zbase:
          XX0 = e. verts[0].co.x 
          YY0 = e. verts[0].co.y
          XX1 = e. verts[1].co.x
          YY1 = e. verts[1].co.y
          if XX0 > XX1: XX0,YY0,XX1,YY1 = XX1,YY1,XX0,YY0
          kk = 999 if YY0 == YY1 else (XX1-XX0)/(YY1-YY0)
          if abs(kk + k) < 0.1:                        # 棟の方向と直角か判定
            elength = e.calc_length() / 2
            if Wshort < elength: Wshort = elength      # Wshort建物半幅(短)
      Sz = 0.4 * Wshort / (Ztop-Zroof)                 # 屋根勾配の補正係数
      Ztop = 0.4 * Wshort + Zroof                      # 勾配修正後の棟高
      #
      ########　屋根を4寸勾配(4:10勾配)に修正する
      obj.select_set(True)                           # 最初に対象の建物のobjを選択
      bpy.ops.mesh.select_all(action="DESELECT")     # そのﾒｯｼｭの全要素を一旦非選択
      meshdata = bmesh.from_edit_mesh(obj.data) 	 # 念のためﾒｯｼｭﾃﾞｰﾀを更新する
      for v in meshdata.verts:
        if v.co.z >= Zroof: v.select_set(True)       # Zroofから上の点を選択する
      bpy.ops.transform.resize(value = (1,1,Sz),
                               center_override = (Xctr,Yctr,Zroof))
      ####### 屋根部を分離し、棟高中心に1.13倍し、ひさしを出す
      move_obj_center(Xctr,Yctr,Ztop)                # 西端のZbaseの高さに原点移動
      bpy.ops.mesh.select_all(action='DESELECT')     # ﾒｯｼｭの要素を全て非選択にする
      for f in meshdata.faces:
        if f.material_index == 1: f.select_set(True) # 屋根の面を選択
      bpy.ops.mesh.split()                           # 屋根の部分を壁から切り離す
      bpy.ops.transform.resize(value=(1.13, 1.13, 1.13),
                               center_override=(Xctr,Yctr,Ztop))
      #
    else:                                  # Bldg-M,Bldg-Hの時,屋根を平にする
      for v in meshdata.verts:
        if v.co.z > Zbase: v.co.z = Ztop             # 屋根の面を Htop の高さに平にする
  #
  ####### 原点を左下に移動し、棟方向を水平(東西)方向にし、UVMapを作成。
  move_obj_center(X0,Y0,Zbase)                      # 西端のZbaseの高さに原点移動
  meshdata = bmesh.from_edit_mesh(obj.data)         # メッシュデータを取得
  bpy.ops.mesh.select_all(action="SELECT")          # ﾒｯｼｭ全ての要素を選択
  bmesh.ops.rotate(
    meshdata,
    verts=meshdata.verts,
    cent=(0.0, 0.0, 0.0),
    matrix=mathutils.Matrix.Rotation(-Zang, 3, 'Z'))
  #
  ###### オブジェクトに応じてUV設定　　##################################################
  if name_obj_main == "Bldg-S":                     # Bldg-S の時のUV設定
    targetuvlayer = meshdata.loops.layers.uv.get(UV_NAME)
    RPnum = [n for n in range(0,Rnum)]
    RPweight = [RP[n][6] for n in range(0,Rnum)] 
    RPsel = random.choices(RPnum, weights = RPweight, k=1)
    WPnum = [n for n in range(0,Wnum)]
    WPweight = [WP[n][6] for n in range(0,Wnum)] 
    WPsel = random.choices(WPnum, weights = WPweight, k=1)
    #
    ku = round((Wlong*WP[WPsel[0]][0]-WP[WPsel[0]][6]) / WP[WPsel[0]][4])
    hoseiu = (ku * WP[WPsel[0]][4] + WP[WPsel[0]][6]) / (Wlong*WP[WPsel[0]][0])
    kv = round((Htop*WP[WPsel[0]][1]-WP[WPsel[0]][7]) / WP[WPsel[0]][5])
    hoseiv = (kv * WP[WPsel[0]][5] + WP[WPsel[0]][7]) / (Htop*WP[WPsel[0]][1])
    for face in meshdata.faces:
      for loop in face.loops:
        loop_uv = loop[targetuvlayer]               # UVレイヤーをUVMapとする
        if face.material_index == 1:                # 屋根面の場合
          uvpos_u = loop.vert.co.x                  # U値(m)
          uvpos_v = -abs(loop.vert.co.y)            # V値(m)
          if uvpos_v > 0: uvpos_v = -uvpos_v        # 範囲外に出ない様に調整
          if uvpos_v < -34: uvpos_v = -34           # 範囲外に出ない様に調整
          uvpos_u = RP[RPsel[0]][0] * uvpos_u + RP[RPsel[0]][2] # U値(dot)
          uvpos_v = RP[RPsel[0]][1] * uvpos_v + RP[RPsel[0]][3] # V値(dot)
        #
        else:                                       # 壁面の場合
          uvpos_u = round(hoseiu*loop.vert.co.x*WP[WPsel[0]][0]/WP[WPsel[0]][4])*WP[WPsel[0]][4]
          if uvpos_u == 0: uvpos_u = 0
          else: uvpos_u = uvpos_u + WP[WPsel[0]][6]+ WP[WPsel[0]][2] # dot
        #  uvpos_u=hoseiu*loop.vert.co.x*WP[WPsel[0]][0]+WP[WPsel[0]][2] # dot
          uvpos_v=hoseiv*loop.vert.co.z*WP[WPsel[0]][1]+WP[WPsel[0]][3] # dot
        loop_uv.uv[0] = uvpos_u / TEX_SIZE_H        # U値( 0-1.0)
        loop_uv.uv[1] = uvpos_v / TEX_SIZE_V        # V値( 0-1.0)
  ###############################################################################
  elif name_obj_main == "Bldg-M":     #Bldg-M の時
    targetuvlayer = meshdata.loops.layers.uv.get(UV_NAME)
    WPnum = [n for n in range(0,Wnum)]
    WPweight = [WP[n][8] for n in range(0,Wnum)]     # 各パターンの選択の重みを指定
    WPsel = random.choices(WPnum, weights = WPweight, k=1) #選択パターン番号をWPsel
    #
    ku = round((Wlong*WP[WPsel[0]][0]-WP[WPsel[0]][6]) / WP[WPsel[0]][4])
    hoseiu = (ku * WP[WPsel[0]][4] + WP[WPsel[0]][6]) / (Wlong*WP[WPsel[0]][0])
    kv = round((Htop*WP[WPsel[0]][1]-WP[WPsel[0]][7]) / WP[WPsel[0]][5])
    hoseiv = (kv * WP[WPsel[0]][5] + WP[WPsel[0]][7]) / (Htop*WP[WPsel[0]][1])
    for face in meshdata.faces:
      for loop in face.loops:
        loop_uv = loop[targetuvlayer]                # UVレイヤーをUVMapとする
        if face.material_index == 1:                 # 屋上面の場合
          uvpos_u = RP[0] * loop.vert.co.x + RP[2]   # U値(dot)
          uvpos_v = RP[1] * loop.vert.co.y + RP[3]   # V値(dot)
          if uvpos_v > RP[4]: uvpos_v = RP[4]        # 範囲外に出ない様に調整
          if uvpos_v < RP[5]: uvpos_v = RP[5]        # 範囲外に出ない様に調整
        else:                                        # 壁面の場合
          uvpos_u = round(hoseiu*loop.vert.co.x*WP[WPsel[0]][0]/WP[WPsel[0]][4])*WP[WPsel[0]][4]
          if uvpos_u == 0: uvpos_u = 0
          else: uvpos_u = uvpos_u + WP[WPsel[0]][6]+ WP[WPsel[0]][2] # dot
          uvpos_v = hoseiv*loop.vert.co.z*WP[WPsel[0]][1] + WP[WPsel[0]][3] # dot
        loop_uv.uv[0] = uvpos_u / TEX_SIZE_H         # U値(係数0-1.0)
        loop_uv.uv[1] = uvpos_v / TEX_SIZE_V         # V値(係数0-1.0)
  #################################################################################
  elif name_obj_main=="Bldg-H":                      #Bldg-H の時
    targetuvlayer = meshdata.loops.layers.uv.get(UV_NAME)
    WPnum = [n for n in range(0,Wnum)]
    WPweight = [WP[n][8] for n in range(0,Wnum)]     # 各パターンの選択の重みを指定
    WPsel = random.choices(WPnum, weights = WPweight, k=1) #選択パターン番号をWPsel
    #
    ku = round((Wlong*WP[WPsel[0]][0]-WP[WPsel[0]][6]) / WP[WPsel[0]][4])
    hoseiu = (ku * WP[WPsel[0]][4] + WP[WPsel[0]][6]) / (Wlong*WP[WPsel[0]][0])
    kv = round((Htop*WP[WPsel[0]][1]-WP[WPsel[0]][7]) / WP[WPsel[0]][5])
    hoseiv = (kv * WP[WPsel[0]][5] + WP[WPsel[0]][7]) / (Htop*WP[WPsel[0]][1])
    for face in meshdata.faces:
      for loop in face.loops:
        loop_uv = loop[targetuvlayer]                # UVレイヤーをUVMapとする
        if face.material_index == 1:                 # 屋上面の場合
          uvpos_u = RP[0] * loop.vert.co.x + RP[2]   # U値(dot)
          uvpos_v = RP[1] * loop.vert.co.y + RP[3]   # V値(dot)
          if uvpos_u < RP[4]: uvpos_u = RP[4]        # 範囲外に出ない様に調整
          if uvpos_u > RP[5]: uvpos_u = RP[5]        # 範囲外に出ない様に調整
        else:                                        # 壁面の場合
          uvpos_u = round(hoseiu*loop.vert.co.x*WP[WPsel[0]][0]/WP[WPsel[0]][4])*WP[WPsel[0]][4]
          if uvpos_u == 0: uvpos_u = WP[WPsel[0]][2]
          else: uvpos_u = uvpos_u + WP[WPsel[0]][6]+ WP[WPsel[0]][2] # dot
          uvpos_v = round(hoseiu*loop.vert.co.z*WP[WPsel[0]][1]/WP[WPsel[0]][5])*WP[WPsel[0]][5]
          if uvpos_v == 0: uvpos_v = WP[WPsel[0]][3]
          else: uvpos_v = uvpos_v + WP[WPsel[0]][7]+ WP[WPsel[0]][3] # dot
        loop_uv.uv[0] = uvpos_u / TEX_SIZE_H         # U値(係数0-1.0)
        loop_uv.uv[1] = uvpos_v / TEX_SIZE_V         # V値(係数0-1.0) 
  #
  else:
    print("Other than Bldg-H, Bldg-M or Bldg-S selected.")
    print()
    break
  #
  ####### オブジェクトのz軸回転を元に戻して、原点を(0,0,0)に戻す。
  bmesh.ops.rotate(
    meshdata,
    verts=meshdata.verts,
    cent=(0.0, 0.0, 0.0),
    matrix=mathutils.Matrix.Rotation(Zang, 3, 'Z'))
  move_obj_center(0,0,0)
  #
  bpy.ops.object.mode_set(mode='OBJECT')             # オブジェクトモードにする
  bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
  obj.select_set(False)
  #
  print("\r", nn, end="")
  nn += 1
now = datetime.datetime.now()
print()
print("Objects joining starts at", now)
for obj_name in selected_name_list:
  bpy.data.objects[obj_name].select_set(True)        # 全オブジェクトを選択する
bpy.ops.object.join()
now = datetime.datetime.now()
print("All UV processing done at", now)