# UV_for_Traffic_01                                                     By Tanzai
#
# 本ソースコードは GPLv3 のライセンスが適用されています。
#
# 選択中の道路オブジェクトに規定のルールでテクスチャを貼ります。
# 詳細は XP11用OSMシーナリー作成　02版 の4.2項を参照。
#
import bpy
import math
import copy
import bmesh
import mathutils
import datetime
import numpy as np
from mathutils import Matrix,Vector
#
#########  サブプログラム　　線のX軸からの角度[西端を(x0, y0)とする]
def z_ang(check_edge):
    X0 = check_edge.verts[0].co.x
    Y0 = check_edge.verts[0].co.y
    X1 = check_edge.verts[1].co.x
    Y1 = check_edge.verts[1].co.y
    if X0 > X1:
        X2 = X0
        Y2 = Y0
        X0 = X1
        Y0 = Y1
        X1 = X2
        Y1 = Y2
    Wlong = math.sqrt((X0-X1)**2 + (Y0-Y1)**2) # 線の長さ
    #
    if X0 == X1:
        Zang = 1.57                            # 傾き角90°
    else:
        k = (Y1-Y0)/(X1-X0)                    # 棟の向き(東を0)
        Zang = math.atan(k)                    # 傾き(東0°,-90°～90°)
    return Zang
#
######### メインプログラム ###############################
now = datetime.datetime.now()
print()
print("UV setting starts at", now)             # 開始時刻を印刷
selected_name_list = [obj.name for obj in bpy.context.selected_objects]
for obj in bpy.context.selected_objects:
    obj.select_set(False)                 # 一旦オブジェクトを全て非選択にする。

####### 以下、個々の道路についてとUV設定を行う
for obj_name in selected_name_list:
    obj = bpy.data.objects[obj_name]               # 対象道路をobjと呼ぶ
    obj.select_set(True)                           # その道路を選択する
    bpy.context.view_layer.objects.active = obj    # その道路をActiveにする
    if obj.type != 'MESH':
        print("WARNING: Obj is not MESH.")
        break                             # その道路がメッシュでない場合は処理しない
    #
    ##########  画像パラメータ設定　　#######################################
    UV_NAME ="UVMap"               # UVを書き込むUVマップ名称、ディフォルトは "UVMap"
    TEX_SIZE_H = 512                                # 画像の水平サイズ(dot)
    TEX_SIZE_V = 512                                # 画像の垂直サイズ(dot)
    #
    # Hrezo:H解像(dot/m)　Vrezo:V解像(dot/m)　Hp:H原点(dot)　Vp:V原点(dot)
    if obj_name[0:5] == "Lane3":              # 先頭5文字が"Lane3"の場合
        Hrezo=11.1
        Vrezo=11.1
        Hp=256
        Vp=51
    elif obj_name[0:5] == "Lane2":            # 先頭5文字が"Lane2"の場合
        Hrezo=9.57
        Vrezo=9.57
        Hp=256
        Vp=132
    elif obj_name[0:5] == "Lane1":            # 先頭5文字が"Lane1"の場合
        Hrezo=10
        Vrezo=10
        Hp=256
        Vp=204
    elif obj_name[0:5] == "Paths":            # 先頭5文字が"Paths"の場合
        Hrezo=10
        Vrezo=10
        Hp=256
        Vp=277
    elif obj_name[0:5] == "Areas":            # 先頭5文字が"Areas"の場合
        Hrezo=1
        Vrezo=1
        Hp=256
        Vp=277
    else :                                    # 鉄道、"Railway"の場合
        Hrezo=6.97
        Vrezo=6.97
        Hp=256
        Vp=350
    #
    ########  UV setting
    bpy.ops.object.mode_set(mode='EDIT')      # 編集モ－ドにする
    meshdata = bmesh.from_edit_mesh(obj.data) # メッシュデータを取得
    targetuvlayer = meshdata.loops.layers.uv.get(UV_NAME)
    #
    
    for face in meshdata.faces:
        Xctr = 0
        Yctr = 0
        for v in face.verts:
            Xctr += v.co.x
            Yctr += v.co.y
        Xctr = Xctr/len(face.verts)
        Yctr = Yctr/len(face.verts)
        #
        if obj_name[0:5] != "Areas":              # 先頭5文字が"Areas"の場合はスキップ
            Eang = []
            Elen = []
            for e in face.edges:
                Eang.append(z_ang(e))
                Elen.append(e.calc_length())
            #
            pare1 = False
            pare2 = False
            if abs(Eang[0] - Eang[1]) < 0.001:
                Zang1 = Eang[0]
                len1 = Elen[0]
                pare1 = True
                if abs(Eang[2] - Eang[3]) < 0.001:
                    Zang2 = Eang[2]
                    len2 = Elen[2]
                    pare2 = True
            if abs(Eang[1] - Eang[2]) < 0.001:
                Zang1 = Eang[1]
                len1 = Elen[1]
                pare1 = True
                if abs(Eang[3] - Eang[0]) < 0.001:
                    Zang2 = Eang[3]
                    len2 = Elen[3]
                    pare2 = True
            if abs(Eang[0] - Eang[2]) < 0.001:
                Zang1 = Eang[0]
                len1 = Elen[0]
                pare1 = True
                if abs(Eang[3] - Eang[1]) < 0.001:
                    Zang2 = Eang[3]
                    len2 = Elen[3]
                    pare2 = True
            if abs(Eang[2] - Eang[3]) < 0.001:
                Zang1 = Eang[2]
                len1 = Elen[2]
                pare1 = True
                if abs(Eang[0] - Eang[1]) < 0.001:
                    Zang2 = Eang[0]
                    len2 = Elen[0]
                    pare2 = True
            if abs(Eang[3] - Eang[0]) < 0.001:
                Zang1 = Eang[3]
                len1 = Elen[3]
                pare1 = True
                if abs(Eang[1] - Eang[2]) < 0.001:
                    Zang2 = Eang[1]
                    len2 = Elen[1]
                    pare2 = True
            if abs(Eang[3] - Eang[1]) < 0.001:
                Zang1 = Eang[3]
                len1 = Elen[3]
                pare1 = True
                if abs(Eang[0] - Eang[2]) < 0.001:
                    Zang2 = Eang[0]
                    len2 = Elen[0]
                    pare2 = True
            #
            if pare1:
                Zang = Zang1
                if pare2:
                    if len1 > len2:
                        Zang = Zang1
                    else:
                        Zang = Zang2
            if pare2:
                Zang = Zang2
        #
        for loop in face.loops:
            loop_uv = loop[targetuvlayer]          # UVレイヤーをUVMapとする
            u1 = loop.vert.co.x - Xctr             # 面の中央値からのU値(m)
            v1 = loop.vert.co.y - Yctr             # 面の中央値からのV値(m)
            if obj_name[0:5] == "Areas":
                uvpos_u = u1                       # 回転なしのU値(m)
                uvpos_v = v1                       # 回転なし	のV値(m)
            else:
                uvpos_u = u1*math.cos(-Zang) - v1*math.sin(-Zang) # -Zang回転後のU値(m)
                uvpos_v = u1*math.sin(-Zang) + v1*math.cos(-Zang) # -Zang回転後のV値(m)
            uvpos_u = Hrezo * uvpos_u + Hp         # U値(dot)
            uvpos_v = Vrezo * uvpos_v + Vp         # V値(dot)
            loop_uv.uv[0] = uvpos_u / TEX_SIZE_H   # U値(係数0-1.0)
            loop_uv.uv[1] = uvpos_v / TEX_SIZE_V   # V値(係数0-1.0)
    #
    bpy.ops.object.mode_set(mode='OBJECT')         # オブジェクトモードにする
    bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
    obj.select_set(False)
    #
now = datetime.datetime.now()
print("All UV processing done at", now)